/* 
 * Copyright (c) 2007 Barath Raghavan.
 * Copyright (c) 2007 The Regents of the University of California.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University of California nor the names of its 
 *    contributors may be used to endorse or promote products derived from 
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <arpa/inet.h>

#ifndef WORDS_BIGENDIAN
#include <byteswap.h>
#endif

#include <stdint.h>
#include <openssl/aes.h>
#include <openssl/rand.h>

#define KEYLEN AES_BLOCK_SIZE // bytes

/* Does a 64-bit equivalent of ntohl */
uint64_t ntohll(uint64_t netlong) {
#ifndef WORDS_BIGENDIAN
  return bswap_64(netlong);
#else
  return netlong;
#endif
}

/* Does a 64-bit equivalent of htonl */
uint64_t htonll(uint64_t hostlong) {
  return ntohll(hostlong);
}

void xor_blocks(unsigned char *out, unsigned char *in1, unsigned char *in2) {
  for (int i = 0; i < AES_BLOCK_SIZE; i++) {
    out[i] = in1[i] ^ in2[i];
  }
}

void print_block(unsigned char *b) {
  unsigned int *b2 = (unsigned int *) b;
  fprintf(stderr, "%x%x%x%x\n", b2[0], b2[1], b2[2], b2[3]);
}

int main(int argc, char *argv[]) {
  if (argc != 2) {
    printf("Usage: %s [missing keybytes] < indata > outdata\n", argv[0]);
    exit(1);
  }

  uint32_t nummasked = atoi(argv[1]);
  if (nummasked != 0 && nummasked != 2 && nummasked != 4) {
    fprintf(stderr, "Error: number of masked keybytes must be 0, 2, or 4.\n");
    exit(1);
  }

  /* generate random key */
  unsigned char keybuf[KEYLEN];
  if (RAND_bytes(keybuf, KEYLEN) == 0) {
    return -1;
  }

  AES_KEY key;
  AES_set_encrypt_key(keybuf, AES_BLOCK_SIZE * 8, &key);

  /* file format: E_k(0), AONT of indata, s', totallen, nummasked */
  unsigned char zeroin[AES_BLOCK_SIZE];
  unsigned char zeroout[AES_BLOCK_SIZE];
  memset(zeroin, 0, AES_BLOCK_SIZE);
  
  AES_encrypt(zeroin, zeroout, &key);
  if (fwrite(zeroout, AES_BLOCK_SIZE, 1, stdout) != 1) {
    perror("fwrite");
    exit(1);
  }

  AES_KEY zerokey;
  AES_set_encrypt_key(zeroin, AES_BLOCK_SIZE * 8, &zerokey);

  /* read hash key block */
  unsigned char hashblock[AES_BLOCK_SIZE];
  memset(hashblock, 0, AES_BLOCK_SIZE);

  /* begin encryption */
  uint64_t i = 1;
  uint64_t totallen = 0;

  unsigned char ctrptext[AES_BLOCK_SIZE];
  memset(ctrptext, 0, AES_BLOCK_SIZE);

  while (!feof(stdin)) {
    unsigned char ptext[AES_BLOCK_SIZE];
    memset(ptext, 0, AES_BLOCK_SIZE);

    /* read data */
    int readlen;
    if ((readlen = fread(ptext, 1, AES_BLOCK_SIZE, stdin)) <= 0) {
      if (ferror(stdin)) {
	perror("fread");
	exit(1);
      }
    }
    
    if (readlen <= 0) {
      break;
    }

    totallen += readlen;

    unsigned char ctrctext[AES_BLOCK_SIZE];
    unsigned char ctext[AES_BLOCK_SIZE];

    uint64_t nctr = htonll(i);
    memcpy(ctrptext, &nctr, sizeof(nctr));

    /* encrypt counter */
    AES_encrypt(ctrptext, ctrctext, &key);

    /* xor with plaintext */
    xor_blocks(ctext, ptext, ctrctext);

    /* output */
    if (fwrite(ctext, AES_BLOCK_SIZE, 1, stdout) != 1) {
      perror("fwrite");
      exit(1);
    }

    /* hash */
    unsigned char hashctext[AES_BLOCK_SIZE];
    
    xor_blocks(ctext, ctext, ctrptext);
    AES_encrypt(ctext, hashctext, &zerokey);

    xor_blocks(hashblock, hashblock, hashctext);

    i++;
  }

  /* hash block xored with key for recovery */
  if (nummasked > 0) {
    memset(keybuf, 0, nummasked);
  }
  xor_blocks(hashblock, hashblock, keybuf);

  if (fwrite(hashblock, AES_BLOCK_SIZE, 1, stdout) != 1) {
    perror("fwrite");
    exit(1);
  }

  /* total length */
  uint64_t ntotallen = htonll(totallen);
  if (fwrite(&ntotallen, sizeof(uint64_t), 1, stdout) != 1) {
    perror("fwrite");
    exit(1);
  }

  /* num masked */
  uint32_t nnummasked = htonl(nummasked);
  if (fwrite(&nnummasked, sizeof(uint32_t), 1, stdout) != 1) {
    perror("fwrite");
    exit(1);
  }

  fflush(stdout);

  return 0;
}
